/*******************************************************************/
/**    MAC 110 - Introdução à Computação                          **/
/**    IME-USP - Primeiro Semestre de 2011                        **/
/**    Turma: 045 - Professor: Fábio Kon                          **/
/**                                                               **/
/**    Terceiro Exercício-Programa -- ByteMaze                    **/
/**    Arquivo: Level.java                                        **/
/**                                                               **/
/**    Rafael Marinaro Verona            7577323                  **/
/**    Ricardo Ferreira Guimarães        7577650                  **/
/**                                                               **/
/*******************************************************************/

import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.io.*;
import java.awt.Color;

import java.net.URL;
import java.util.*;

/**
 * The class Level control the actors location, menu, and events between maps.
 * 
 * @author Ricardo F. Guimarães and Rafael M. Verona
 * @version 1.8
 */
public class Level  extends World
{
    private final int TILE_SIZE = 50; //Tamanho do tile (quadrado) do mapa
    
    private GreenfootSound bgm =  new GreenfootSound("Rocket.mp3"); //Música de fundo
    
    private final int PLAYER_HP = 90; //Vida máxima do jogador
    private final int PLAYER_ENERGY = 35; //Energia máxima do jogador
    
    private Player p = new Player(PLAYER_HP,PLAYER_ENERGY); //Objeto da classe player
    
    private int playerColumn; //Coluna onde está o player
    private int playerLine; //Linha onde está o player
     
    private int level = 0; //Inidica em qual nível o usuário está
    private final int lastLevel = 15; //Indica qual é a última fase
    
    //Elemento da interface (HUD)
    private Panel radar = new Panel(); //Indica número de inimigos ao redor
    private Panel scoreBoard = new Panel(); //Indica a pontuação atual
    private Panel messagePanel = new Panel(); //Mostra mensagens sobre o jogo
    private Panel label = new Panel(); //Rótulo das barras de vida (HP) e energia (EP)
    private Panel inventory = new Panel(); //Rótulo da áre utilizada como indicador de senhas
    
    private Bar hpBar = new Bar(PLAYER_HP,Color.RED); //Barra de vida
    private Bar epBar = new Bar(PLAYER_ENERGY, Color.BLUE); //Barra de energia
    //
    
    private int acumulatedScore = 0; //Pontos acumulados
    private int numberOfSources = 0; //Número de códigos-fonte
    
    //Botões para o menu
    Button playButton = new Button("buttonPlay.png"); //Botão "Jogar"
    Button helpButton = new Button("buttonHelp.png"); //Botão "Ajuda"
    Button creditsButton = new Button("buttonCredits.png"); //Botão "Créditos"
    //
    
    private int fullMap[][] = new int [12][16]; //Matriz que será usada na criação dos níveis
    
    /**
     * Constructor for objects of class Level.
     */    
    public Level()
    {  
        super(800, 600, 1,false); //Cria um novo Mundo com 800 x 600 pixels, e células com 1 pixel de tamanho
        
        setPaintOrder(Outdoor.class, Explosion.class,Button.class, Panel.class, Bar.class, Player.class, SourceCode.class, Password.class, Wall.class, 
                      Enemy.class, Door.class, Central.class, Floor.class); //Define a ordem de sobreposição (desenho) das classes
                      
        Greenfoot.setSpeed(50); //Define a velocidade do mundo
        
        drawMainMenu(); //Desenha o menu principal
    }
    
    /**
     * Draw the main menu on the screen
     */
    private void drawMainMenu()
    {
        setBackground("mainTitle.png"); //Coloca a imagem de fundo
        addObject(playButton, 395, 190); //Coloca o botão "PLAY"
        addObject(helpButton, 395, 300); //Coloca o botão "HELP"
        addObject(creditsButton, 395, 410); //Coloca o botão "CREDITS"
    }
    
    /**
     * Act - do whatever the Level wants to do. This method is called whenever
     * the 'Act' or 'Run' button gets pressed in the environment.
     */
    public void act()
    {
        if (level == 0) //Se p level for 0, o jogo ainda não começou
        {
            menu();
        }
    }
   
    
    /**
     * Load the file specified on the matriz fullMap.
     * @param fileName the name of the file to be loaded
     */
    private void loadMatrix(String fileName)
    {
        BufferedReader bufferedReader = null; //Cria um Buffered Reader
        
        try {
            URL url = getClass().getClassLoader().getResource(fileName); //Cria uma nov URL
            bufferedReader = new BufferedReader( new InputStreamReader (url.openStream()) ); //Inicializa o Buffered Reader com uma stream para o arquivo
            
            int symbol = -1; //Símbolo lido
            int i = 0, j = 0; //Indices da matriz
            
            while ((symbol = bufferedReader.read()) != -1) //Enquanto não for fim de arquivo
            {
                if (symbol >= 'A' && symbol <= 'z') //Se for um caracter válido
                {
                    fullMap[i][j] = symbol; //Armazena na matriz
                    j++;
                    if (j == 16)
                    {
                        j = 0;
                        i++;
                    }
                }
            }
            
        } catch (FileNotFoundException ex) {
            ex.printStackTrace();
        } catch (IOException ex) {
            ex.printStackTrace();
        } finally {
            //Close the BufferedReader
            try {
                if (bufferedReader != null)
                    bufferedReader.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
      }
       
    /**
     * Prepare the world for the start of the program. That is: create the initial
     * objects and add them to the world.
     */
    private void prepare()
    {
        numberOfSources = 0;
        for (int i = 0; i < fullMap.length; i++) //Percorre a matriz
            for (int j = 0; j < fullMap[0].length; j++)
            {
                Actor x;
                if (fullMap[i][j] == 'w') //Cria uma parede
                {
                    x = new Wall("Wall.jpg");
                }
                else if (fullMap[i][j] == 't') //Cria um terminal e define a posição do Player
                {
                    x = new Floor("terminal.png");
                    playerColumn = (j * TILE_SIZE) + TILE_SIZE/2;
                    playerLine = (i * TILE_SIZE) + TILE_SIZE/2;
                }    
                else if (fullMap[i][j] == 'f') //Cria um piso
                {    
                    x = new Floor("floor.png");
                }    
                else if (fullMap[i][j] >= 'b' && fullMap[i][j] <= 'e') //Cria um inimigo definindo seu dano
                {    
                    x = new Enemy( (int) Math.pow(2, fullMap[i][j] - 'a'));
                }
                else if (fullMap[i][j] == 'x') //Cria uma central
                {    
                    x = new Central();
                }
                else if (fullMap[i][j] >= 'i' && fullMap[i][j] < 'o') //Cria uma senha com seu valor
                {
                    x = new Password(fullMap[i][j] - 'i', colorSelect(fullMap[i][j] - 'i'));
                }
                else if (fullMap[i][j] >= 'I' && fullMap[i][j] < 'O') //Cria uma porta com seu valor
                {
                    x = new Door(fullMap[i][j] - 'I', colorSelect(fullMap[i][j] - 'I'));
                }
                else if (fullMap[i][j] == 's') //Cria um código-fonte e incrementa o contador
                {
                    x = new SourceCode();
                    numberOfSources++;
                }
                else //Senão cria um novo piso
                {    
                    x = new Floor("floor.png");
                }
                addObject(x,(j * TILE_SIZE) + TILE_SIZE/2,(i * TILE_SIZE) + TILE_SIZE/2); //Efetivamente adiciona o objeto no mundo
            }
            
            p.setNumberOfSources(numberOfSources); //Define o número de códigos-fonte do nível para o Player
        }
    
    /**
     * Draws the interface (HUD) on the screen
     */
    private void drawHUD()
    {
        addObject(radar, 200, 25);
        addObject(scoreBoard, 400, 25);
        addObject(hpBar, 200, 560);
        addObject(epBar, 200, 580);
        addObject(label, 28, 570);
        addObject(inventory, 400, 570);
        addObject(messagePanel,600,25);
        messagePanel.update("Jacked in Zone " + level);
        label.update("HP\nEP");
        inventory.update("KEYS: ");
    }
    
    /**
     * Returns the Panel used for radar (enemies around)
     * @see Panel
     */
    public Panel radarPanel()
    {
        return radar;
    }
    
    /**
     * Returns the Panel used for score (points)
     * @see Panel
     */
    public Panel scorePanel()
    {
        return scoreBoard;
    }
    
    /**
     * Returns the Panel used for messages in game
     * @see Panel
     */
    public Panel messageBox()
    {
        return messagePanel;
    }
    
    /**
     * Returns the Bar used as Life Indicator
     * @see Bar
     */
    public Bar lifeIndicator()
    {
        return hpBar;
    }
    
    /**
     * Returns the Bar used as Energy Indicator
     * @see Bar
     */
    public Bar energyIndicator()
    {
        return epBar;
    }
    
    /**
     * Plays the background music in loop
     */
    public void started()
    {
       bgm.playLoop();
    }
    
    /**
    * Stops the background music
    */
    public void stopped()
    {
       bgm.stop();
    }
    
    /**
    * Advance the game to the next level or end the game if it was int the last level
    * @param partialScore points aquired until now
    */
    public void nextLevel(int partialScore)
    {
        setBackground("floor.png"); //Coloca uma outra imagem de fundo
        
        if (level < lastLevel) //Se não estava no último nível
        {
            level++; //Aumenta o indicador de nível
            removeObjects(getObjects(Actor.class)); //Limpa o mundo
            
            if (level == 1) //Se é o primeiro nível
                addObject(new Outdoor("history.png",true), 400,300); //Mostra a tela com uma pequena história
            else //Senão
                addObject(new Outdoor("nextLevel.png",true), 400,300); //Mostra uma imagem informando que passou de nível 
                
            loadMatrix("level" + level + ".txt"); //Carrega o arquivo do respectivo nível
            prepare(); //Prepara o mundo
            drawHUD(); //Desenha a interface
            acumulatedScore = partialScore; //Define a pontuação acumulada
            p.setScore(acumulatedScore); //Passa para o player a pontuação atual
        }
        else //Se era o último nível
        {
            addObject(new Outdoor("endGame.png",false), 400,300); //Mostra a tela de fim de jogo
            Greenfoot.stop(); //Pára a execução
        }
    }
    
    /**
    * Restart the actual level if the player has enough points, else shows Game Over
    * @param partialScore points aquired until now
    */
    public void sameLevel(int partialScore)
    {
        removeObjects(getObjects(Actor.class)); //Limpa o mundo
        
        if (partialScore <= 1500) //Se a pontuação é menor ou igual a 2000
        {
            lostGame(); //Game Over
        }
        else
        {
            continueGame(partialScore - 1500); // O jogo continua no mesmo nível
        }
    }
    
    /**
    * Used in case of Game Over.
    */
    private void lostGame()
    {
        level = 0; //Reinicia o jogo do começo
        addObject(new Outdoor("erased.png",false), 400,300); //Mostra a tela de Game Over
        p = new Player(PLAYER_HP,PLAYER_ENERGY); //Cria um novo player
        p.setScore(0); //Inicia potação do player com zero
        acumulatedScore = 0; //Zera toda a pontuação
        drawMainMenu(); //Desenha o menu principal
        menu(); //Volta para o menu
    }
    
    /**
    * Used in case of death of the player, if it can continue the game.
    * @param value points aquired
    */
    private void continueGame(int value)
    {
        addObject(new Outdoor("died.png",true), 400,300); //Mostra a tela de morete
        p = new Player(PLAYER_HP,PLAYER_ENERGY); //Cria um novo player
        p.setScore(value); //Define a pontuação
        prepare(); //Prepara novamente o mundo
        drawHUD(); //Desenha interface
        acumulatedScore = value; //Define a nova pontuação acumulada
    }
    
    /**
    * Adds the player to the world in the right position
    */
    public void addPlayer()
    {
        Greenfoot.delay(2); //Delay usado para impedir que as teclas pressionadas antes do jogo iniciar interfiram no player
        addObject(p, playerColumn, playerLine); //Adiciona o Player ao mundo
    }
    
    /**
    * Provides easy acess to colors define for doors and keys
    * @param index code for the color that will be used
    * @returns the respective color for the index
    */
    private java.awt.Color colorSelect(int index)
    {
         if (index == 0)
            return Color.RED;
         else if (index == 1)
            return Color.BLUE;
         else if (index == 2)
            return Color.GREEN;
         else if (index == 3)
            return Color.YELLOW;
         else if (index == 4)
            return Color.CYAN;
         else if (index == 5)
            return Color.WHITE;
         else
            return Color.BLACK;
    }
    
    /**
    * The menu itself, containning all the actions related with the buttons
    */
    private void menu()
    {          
        if (Greenfoot.mouseClicked(helpButton)) //Se o usuário clicar na ajuda
        {
            addObject(new Outdoor("help2.png",false), 400,300); //Adicona as telas de ajuda
            addObject(new Outdoor("help1.png",false), 400,300);
            addObject(new Outdoor("help0.png",false), 400,300);
        }
            
        else if (Greenfoot.mouseClicked(creditsButton)) //Se clicar nos créditos
        {
            addObject(new Outdoor("credits.png",false), 400,300); //Adiciona a tela de cŕeditos
        }
        else if (Greenfoot.mouseClicked(playButton)) //Se clicar em jogar
        {
           nextLevel(0); //Inicia o jogo do começo
        }
    }
}
