/*******************************************************************/
/**    MAC 110 - Introdução à Computação                          **/
/**    IME-USP - Primeiro Semestre de 2011                        **/
/**    Turma: 045 - Professor: Fábio Kon                          **/
/**                                                               **/
/**    Terceiro Exercício-Programa -- ByteMaze                    **/
/**    Arquivo: Player.java                                       **/
/**                                                               **/
/**    Rafael Marinaro Verona            7577323                  **/
/**    Ricardo Ferreira Guimarães        7577650                  **/
/**                                                               **/
/*******************************************************************/

import greenfoot.*;
import java.util.*;

/**
 * Main object, which will react to the input, an then act with the other objects in the world.
 * 
 * @author Ricardo F. Guimarães and Rafael M. Verona
 * @version 1.1
 */
public class Player  extends Actor
{
    private final int RIGHT = 0;    //Constantes para facilitar o uso das direções
    private final int UP = 270;
    private final int LEFT = 180;
    private final int DOWN = 90;
    
    private final int DIST_MOVE = 50;   //Quanto o jogador se move no mínimo a cada vez que ele anda
    private final int SIGHT = 75;       //Distância de alcance do radar de inimigos
    
    private int moveSpeed = 5; //Quanto o jogador se move a cada act
    private int moved = 0; //Quanto o jogador já se moveu
    private boolean moving = false; //Indica se o jogador está ou não se movendo
    private boolean alerted = false; //Indica se o radar já atuou para aquele movimento do Player
    
    private int MAX_LIFE;   //Guarda a vida máxima do jogador
    private int MAX_ENERGY; //Guarda a energia máxima do jogador
    
    private int life;   //Guarda a vida atual do jogador
    private int energy; //Guarda a energia atual do jogador
    private int score;  //Guarda a pontuação atual do jogador
    
    private int sourcesLeft;    //Indica quantos Códigos Fonte ainda faltam para completar o nível
    
    private List<Integer> passwords = new ArrayList<Integer>(); //Lista que guarda as senhas obtidas pelo jogador
    
    /**
     * Creates an object Player with an life and an energy
     * @param hp life
     * @param ep energy
     */
    public Player(int hp, int ep)
    {
        life = MAX_LIFE = hp;
        energy = MAX_ENERGY = ep;
    }
        
    /**
     * Set a Score for the Player
     * @param acumulatedScore points acumulated through the past levels
     * 
     */
    public void setScore(int acumulatedScore)
    {
        score = acumulatedScore;
    }
    
    /**
     * Set the number of Source Codes to be collected on the current level
     * @param numberOfSources number of Source Codes
     * 
     */
    public void setNumberOfSources(int numberOfSources)
    {
        sourcesLeft = numberOfSources;
    }
    
    /**
     * Act - do whatever the Player wants to do. This method is called whenever
     * the 'Act' or 'Run' button gets pressed in the environment.
     */
    public void act() 
    {
        if (!moving) //Se o jogador está parado
        {
            stepInTo(); //Verifica onde ele está
            processDirections(); //Processa as direções
            interactWithObject(); // Interage com os objetos à frente
        }
        else //Se o jogador está se movendo
        {
            if (moved % 10 == 0) //Troca de imagens a cada 10 acts
                    setImage(new GreenfootImage("player" + (moved/10) + ".png"));
            moved += moveSpeed; //Aumenta o quanto já se moveu
            move(moveSpeed); //Faz o player andar
            
            if (moved == DIST_MOVE) //Se já se moveu tudo o que deveria
            {
                setImage(new GreenfootImage("player0.png")); //Coloca a imagem do Player parado
                moving = false; //O Player pára de se mover
                moved = 0; //Reinicia o contador de deslocamento
            }
        }
        
        updateHUD(); //Altera as informações dadas na tela
        
        if (!isAlive()) //Se o Player morreu
        {
            GreenfootSound defeatSfx =  new GreenfootSound("33220__HardPCM__Chip080.mp3");
            defeatSfx.play(); //Reproduz o efeito sonoro de derrota
            
            passwords = new ArrayList<Integer>(); //Cria uma nova lista de senhas
            ((Level) (getWorld()) ).sameLevel(score + energy + life); //Chama um método do mundo (Level) para verificar se pode refazer ou se termina o jogo
        }
        else if (sourcesLeft == 0) //Se o Player está vivo e coletrou todas os Códigos-Fonte do Nível
        {
            GreenfootSound victorySfx =  new GreenfootSound("34232__HardPCM__Chip116.mp3");
            victorySfx.play(); //Reproduz o efeito sonoro de vitória
            
            passwords = new ArrayList<Integer>(); //Cria uma nova lista de senhas
            ((Level) (getWorld()) ).nextLevel(score + energy + life); //Chama um método do mundo (Level) para verificar se avança ou se termina o jogo
        }
    }   
    
    /**
     * It reacts to user input and, according to the current position of the player, 
     * turning or moving the player in the direction of the pressed key.
     */
    private void processDirections()
    {
        if (Greenfoot.isKeyDown("right")) //Se o usuário apertou "DIREITA"
        {
            if (getRotation() != RIGHT) //Se o Player não está virado para direita
                turnPlayer(RIGHT); //O player volta-se para direita
            else //Se o Player está virado para direita
                movePlayer(RIGHT);  //O Player anda para a direita
        }
        
        else if (Greenfoot.isKeyDown("up")) //Se o usuário apertou "CIMA"
        {
            if (getRotation() != UP) //Se o Player não está virado para cima
                turnPlayer(UP); //O player volta-se para cima
            else //Se o Player está virado para cima
                movePlayer(UP); //O Player anda para a cima
        }
        
        else if (Greenfoot.isKeyDown("left")) //Se o Player não está virado para esquerda
        {
            if (getRotation() != LEFT) //O player volta-se para esquerda
                turnPlayer(LEFT); //O player volta-se para esquerda
            else //Se o Player está virado para esquerda
                movePlayer(LEFT); //O Player anda para a esquerda
        }
        
        else if (Greenfoot.isKeyDown("down"))
        {
            if (getRotation() != DOWN) //Se o Player não está virado para baixo
                turnPlayer(DOWN); //O player volta-se para baixo
            else //Se o Player está virado para baixo
                movePlayer(DOWN); //O Player anda para a baixo
        }
         
    }
    
    /**
     * Turns the player progressively to the desired rotation.
     * @param angleFinal desired rotation
     */
    private void turnPlayer(int angleFinal)
    {
        int angleStart = getRotation(); //Ângulo de início
        
        while (getRotation() != angleFinal) //Enquanto a rotação for di
        {
            if (angleStart == RIGHT && angleFinal == UP)
                turn(-9);
            
            else if (angleStart > angleFinal && angleStart <= (angleFinal +180))
                turn(-9);
            else
                turn(+9);
            Greenfoot.delay(1);
        }
    }
    
    /**
     * If the Player can move, then set the status moving true, set that the "radar" should warn the user,
     * and decreases the energy by 1.
     * @param direction direction in which the Player wants to move
     */
    private void movePlayer(int direction)
    {
        if (canMove(direction)) //Se o Player pode se mover
        {
            moving = true; //Ele está se movendo
            alerted = false; //Deve ser alertado pelo radar
            changeEnergy(-1); //Sua energia diminui em 1
          
        }
    }
    
    /**
     * Checks if the player can move, if there is an object that belongs to the class Walkable in his front.
     * @param direction direction in which the Player wants to move
     */
    private boolean canMove(int direction)//Verifica se o player pode se mover, se há um objeto que pertence a classe Walkable a sua frente.
    {
        Walkable walk = null;
        
        if (direction == RIGHT)
            walk = (Walkable) (getOneObjectAtOffset(DIST_MOVE,0,Walkable.class));
        
        else if (direction == LEFT)
            walk = (Walkable) (getOneObjectAtOffset(-DIST_MOVE,0,Walkable.class));
            
        else if (direction == UP)
            walk = (Walkable) (getOneObjectAtOffset(0,-DIST_MOVE,Walkable.class));
            
        else if (direction == DOWN)
            walk = (Walkable) (getOneObjectAtOffset(0,DIST_MOVE,Walkable.class));
            
        if (walk != null)
        {
              return walk.isWalkable();
        }
        else
            return false;
    }
    
    /**
     * Returns the Actor which the player is facing (the Actor in his front).
     */
    private Actor facingActor()
    {
        if (getRotation() == RIGHT)
            return getOneObjectAtOffset(DIST_MOVE,0,Actor.class);
        else if (getRotation() == UP)
            return getOneObjectAtOffset(0,-DIST_MOVE,Actor.class);
        else if (getRotation() == LEFT)
            return getOneObjectAtOffset(-DIST_MOVE,0,Actor.class);
        else if (getRotation() == DOWN)
            return getOneObjectAtOffset(0,DIST_MOVE,Actor.class);
        else
            return null;
    }
    
    /**
     * Method used to interact with the following objects: Enemies, Central and Ports.
     */
    private void interactWithObject()//Método usado para interagir com os seguintes objetos: Inimigos, Centrais e Portas.
    {
        String key = Greenfoot.getKey();
        if (key != null)
            if (key.equals("space"))//Se a tecla precionada for "space"
            {
                Actor a = facingActor();
                
                if ( a instanceof Enemy)//Se a player estiver frente a frente com um inimigo.
                {
                    ( (Enemy) a ).turnVisible();//O inimigo fica visível.
                    if (Greenfoot.getRandomNumber(10) < 7)//O ataque é realizado com 70% de chance de acerto.
                    {
                        addPoints( ( (Enemy) a ).die() );//O inimigo morre e é removido.
                        changeEnergy(5);//Sua vida aumenta em 5 pontos.
                        
                    }
                    else//Se o ataque for malsucedido, 30% de chance de ocorrência.
                    {
                        GreenfootSound missedSfx =  new GreenfootSound("31876__HardPCM__Chip039.mp3");
                        missedSfx.play();//Um som de alerta é executado.
                        takeDamage(( (Enemy) a ).doDamage());//O inimigo provoca uma redução na vida do player, através do método "doDamage".
                    }
                }
                else if ( a instanceof Central)//Se o player estiver frente a frente com uma central.
                {
                    if (Greenfoot.getRandomNumber(10) < 5)//O ataque é realizado com 50% de chance de acerto.
                    {
                        
                        addPoints( ( (Central) a ).disable() );//A central é desativada.
                        changeEnergy(10);//A energia do player aumenta em 10 pontos.
                    }
                    else//Se o ataque for malsucedido, 50% de chance de ocorrência.
                    {
                        GreenfootSound missedSfx =  new GreenfootSound("31876__HardPCM__Chip039.mp3");
                        missedSfx.play();//Um som de alerta é executado.
                        
                        ((Level) (getWorld()) ).messageBox().update("Attacked by Central");//Uma mensagem de alerta aparece na tela (Attacked by Central).
                        takeDamage(8);//E sua vida é reduzida em 8 pontos.
                    }
                }
                else if ( a instanceof Door)//Se o player estiver frente a frente com uma porta.
                {//E se ele possuir a chave correspondente à porta.
                    ((Level) (getWorld()) ).messageBox().update("Working on the Door");//Uma mensagem de alerta aparece na tela (Working on the Door).
                    unlockDoor((Door) a);//A porta abre-se e é removida.
                }
                
                else//Se o player não estiver carregando consigo a chave correspondente à porta.
                {
                    GreenfootSound errorSfx =  new GreenfootSound("35715__HardPCM__Chip133.mp3");
                    errorSfx.play();//Um alerta sonoro é emitido.
                    ((Level) (getWorld()) ).messageBox().update("ERROR");//Uma mensagem de erro é exibida na tela.
                    takeDamage(1);//E a vida do player diminui em 1 ponto.
                }
            }
    }
    
    /**
     * Method that computes the damage to the life of the player.
     * @param value value to be computed
     */
    private void takeDamage (int value) //Método que computa os danos na vida do player.
    {
        changeHealth(-value); //Executa o método "changeHealth" com o valor da vida a ser reduzido.
    }
    
    /**
     * Method that computes the damage or heal to the life of the player.
     * @param value value to be computed
     */
    private void changeHealth (int value)//Método que altera a vida do player.
    {
        life += value;
        
        if (life > MAX_LIFE)
            life = MAX_LIFE;
        else if (life < 0)
            life = 0;
    }
    
    /**
     * Method that computes changes to the energy of the player.
     * @param value value to be computed
     */
    private void changeEnergy(int value)
    {
        energy += value;
        
        if (energy > MAX_ENERGY)
            energy = MAX_ENERGY;
        else if (energy < 0)
            energy = 0;
    }
    
    /**
     * Method that adds points to the player score.
     * @param points points to be added
     */
    private void addPoints(int points)
    {
        score += points;
    }
    
    /**
     * Method that verify if the player is alive.
     */
    public boolean isAlive()//Método que verifica se o player está vivo.
    {
        return (life > 0) && (energy > 0);//O player está vivo (true), se sua vida E sua sua energia são maiores que zero.
    }
    
    /**
     * Method that verify if the player "collided" with something.
     */
    private void stepInTo()
    {
        if (getOneIntersectingObject(Enemy.class) != null) //Se colidiu com inimigo
        {
            GreenfootSound bumpedSfx =  new GreenfootSound("35714__HardPCM__Chip132.mp3");
            bumpedSfx.play(); //Reproduz som de alerta
            
            takeDamage( ( (Enemy) getOneIntersectingObject(Enemy.class) ).doDamage() ); //Diminui a vida do Player
             ( (Enemy) getOneIntersectingObject(Enemy.class) ).turnVisible(); //Torna visível o inimigo
             pushBack(); //Joga o Player para trás
        }
        else if (getOneIntersectingObject(Password.class) != null) //Se "colidiu" com uma senha
        {
            passwords.add( ((Password) (getOneIntersectingObject(Password.class)) ).getKey()  ); //Adiciona a senha à lista
        }
        else if (getOneIntersectingObject(SourceCode.class) != null) //Se "colidiu" com um código-fonte
        {
            changeHealth(7); //Regenera vida
            addPoints( ( (SourceCode) getOneIntersectingObject(SourceCode.class) ).got() ); //Adiciona pontos
            sourcesLeft--; //Diminui o número de códigos - fonte restantes
        }
    }
    
    /**
     * Method pushes the player back
     */
    private void pushBack()
    {
        move(-DIST_MOVE);
        turn(180);
    }
    
    /**
     * Method that add a new password to the list.
     * @param password code
     */
    private void addPassword(int value)
    {
        ((Level) (getWorld()) ).messageBox().update("Password added"); //Envia mensagem
        passwords.add(value); //Adiciona a senha
    }
    
    /**
     * Try to unlock a door using all Passwords avaliable.
     * @param door the door which the player is trying to open
     */
    private void unlockDoor(Door door)
    {
        int i = 0;
        
        while ((i < passwords.size()) && door.tryOpen(passwords.get(i) ) == false) //Enquanto não abriu, nem chegou até o fim
            i++;
            
        if (i !=passwords.size()) //se conseguiu abrir
        {
            GreenfootSound openedDoorSfx =  new GreenfootSound("33221__HardPCM__Chip081.mp3");
            openedDoorSfx.play(); //Reproduz som de sucesso
            
            getWorld().removeObjects(getWorld().getObjectsAt(440 + (25 * passwords.get(i)), 575, Password.class)); //Remove a senha do inventário
            passwords.remove(i); //Remove a senha da lista
        }
        
     }
     
     /**
     * Indicates how many enemies are around the Player
     * @return number of enemies around
     */
     public int enemiesAround()
     {
        List enemies = getObjectsInRange(SIGHT,Enemy.class); //Cria uma lista dos inimigos ao redor
        
        if (enemies.size() > 0 && !alerted) //Se há inimigos e player ainda não foi alertado
        {
            GreenfootSound radarSfx =  new GreenfootSound("33113__HardPCM__Chip069.mp3");
            radarSfx.play(); //Alerta o player
            alerted = true; //Agora ele já foi alertado
        }
         
        return enemies.size(); //Número de inimigos ao redor
     }
     
     /**
     * Update all bars, messege box on the screen which depend on the Player
     */
     private void updateHUD()
     {
        ((Level) (getWorld()) ).radarPanel().update("Enemies: " + enemiesAround());
        ((Level) (getWorld()) ).scorePanel().update("Score: " + score);
        ((Level) (getWorld()) ).lifeIndicator().update(life);
        ((Level) (getWorld()) ).energyIndicator().update(energy);
    }
     
}